import h5py
import numpy as np

import pytest

from helper_methods import retrieve_data

from dclab import new_dataset, rtdc_dataset
from dclab.rtdc_dataset.linker import (
    ExternalDataForbiddenError,
    check_external,
    combine_h5files)


def test_linker_features_base():
    """Create a dataset that refers to a basin in a relative path"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Create complementary dataset
    with h5py.File(h5path, "a") as src, h5py.File(h5path_small, "w") as dst:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=dst,
                               features="scalar")
        del src["/events/deform"]
        # sanity checks
        assert "deform" in dst["events"]
        assert "deform" not in src["events"]

        assert "image" not in dst["events"]
        assert "image" in src["events"]

    fd = combine_h5files([h5path_small, h5path])

    # Now open the scalar dataset and check whether basins are defined
    with new_dataset(fd) as ds:
        assert "image" in ds
        assert "deform" in ds
        assert np.median(ds["image"][0]) == 151


def test_linker_features_error_external_links_dataset():
    """External links are forbidden"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_image = h5path.with_name("image.hdf5")

    # Dataset creation
    with h5py.File(h5path) as src, \
            h5py.File(h5path_image, "w") as h5:
        # write image data to separate file
        h5["image"] = src["/events/image"][:]

    # turn image into an external link
    with h5py.File(h5path, "a") as src:
        del src["/events/image"]
        src["/events/image"] = h5py.ExternalLink(
            str(h5path_image), "image"
        )
        # sanity check
        assert check_external(src) == (True, "/events/image")

    with pytest.raises(ExternalDataForbiddenError,
                       match=r"not permitted for security reasons"):
        combine_h5files([h5path], external="raise")


def test_linker_features_error_external_links_group():
    """External links are forbidden"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_logs = h5path.with_name("logs.hdf5")

    # Dataset creation
    with h5py.File(h5path) as src, \
            h5py.File(h5path_logs, "w") as h5:
        # write image data to separate file
        h5logs = h5.require_group("logs")
        for key in src["logs"]:
            h5logs[key] = src["logs"][key][:]

    # turn image into an external link
    with h5py.File(h5path, "a") as src:
        del src["/logs"]
        src["/logs"] = h5py.ExternalLink(
            str(h5path_logs), "/"
        )
        # sanity check
        assert check_external(src) == (True, "/logs")

    with pytest.raises(ExternalDataForbiddenError,
                       match=r"not permitted for security reasons"):
        combine_h5files([h5path], external="raise")


def test_linker_features_error_original_links_dataset():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_image = h5path.with_name("image.hdf5")

    # Dataset creation
    with h5py.File(h5path) as src, h5py.File(h5path_image, "w") as h5:
        # write image data to separate file
        h5["image"] = src["/events/image"][:]

    # turn image into an external link
    with h5py.File(h5path, "a") as src:
        del src["/events/image"]
        src["/events/image"] = h5py.ExternalLink(
            str(h5path_image), "image"
        )
        # sanity check
        assert check_external(src) == (True, "/events/image")

    with pytest.raises(ExternalDataForbiddenError,
                       match=r"not permitted for security reasons"):
        combine_h5files([h5path], external="raise")


def test_linker_logs():
    h5path = retrieve_data("fmt-hdf5_raw-cytoshot-exported.zip")
    fd = combine_h5files([h5path], external="raise")
    with new_dataset(fd) as ds:
        assert ds.logs
        assert len(ds.logs) == 2
        assert "so2exp_cytoshot-acquisition" in ds.logs


def test_linker_tables():
    h5path = retrieve_data("fmt-hdf5_raw-cytoshot-exported.zip")
    fd = combine_h5files([h5path], external="raise")
    with new_dataset(fd) as ds:
        assert ds.tables
        assert len(ds.tables) == 2
        assert "so2exp_cytoshot_monitor" in ds.tables
        assert np.allclose(
            ds.tables["so2exp_cytoshot_monitor"]["rotation"][0],
            2.0625)



"""
With basins, you can create analysis pipelines that result in output files
which, when opened in dclab, can access features stored in the input file
(without having to write those features to the output file).
"""
from __future__ import annotations

import abc
import threading
from typing import Dict, List, Literal
import weakref

import numpy as np


class BasinmapFeatureMissingError(KeyError):
    """Used when one of the `basinmap` features is not defined"""
    pass


class BasinAvailabilityChecker(threading.Thread):
    """Helper thread for checking basin availability in the background"""
    def __init__(self, basin, *args, **kwargs):
        super(BasinAvailabilityChecker, self).__init__(*args, daemon=True,
                                                       **kwargs)
        self.basin = basin

    def run(self):
        self.basin.is_available()


class Basin(abc.ABC):
    """A basin represents data from an external source

    The external data must be a valid RT-DC dataset, subclasses
    should ensure that the corresponding API is available.
    """
    id_getters = {}

    def __init__(self,
                 location: str,
                 name: str = None,
                 description: str = None,
                 features: List[str] = None,
                 measurement_identifier: str = None,
                 mapping: Literal["same",
                                  "basinmap0",
                                  "basinmap1",
                                  "basinmap2",
                                  "basinmap3",
                                  "basinmap4",
                                  "basinmap5",
                                  "basinmap6",
                                  "basinmap7",
                                  "basinmap8",
                                  "basinmap9",
                                  ] = "same",
                 mapping_referrer: Dict = None,
                 **kwargs):
        """

        Parameters
        ----------
        location: str
            Location of the basin, this can be a path or a URL, depending
            on the implementation of the subclass
        name: str
            Human-readable name of the basin
        description: str
            Lengthy description of the basin
        features: list of str
            List of features this basin provides; This list is enforced,
            even if the basin actually contains more features.
        measurement_identifier: str
            A measurement identifier against which to check the basin.
            If this is set to None (default), there is no certainty
            that the downstream dataset is from the same measurement.
        mapping: str
            Which type of mapping to use. This can be either "same"
            when the event list of the basin is identical to that
            of the dataset defining the basin, or one of the "basinmap"
            features (e.g. "basinmap1") in cases where the dataset consists
            of a subset of the events of the basin dataset. In the latter
            case, the feature defined by `mapping` must be present in the
            dataset and consist of integer-valued indices (starting at 0)
            for the basin dataset.
        mapping_referrer: dict-like
            Dict-like object from which "basinmap" features can be obtained
            in situations where `mapping != "same"`. This can be a simple
            dictionary of numpy arrays or e.g. an instance of
            :class:`.RTDCBase`.
        kwargs:
            Additional keyword arguments passed to the `load_dataset`
            method of the `Basin` subclass.

        .. versionchanged: 0.58.0

            Added the `mapping` keyword argument to support basins
            with a superset of events.
        """
        #: location of the basin (e.g. path or URL)
        self.location = location
        #: user-defined name of the basin
        self.name = name
        #: lengthy description of the basin
        self.description = description
        # features this basin provides
        self._features = features
        #: measurement identifier of the referencing dataset
        self.measurement_identifier = measurement_identifier
        self._measurement_identifier_verified = False
        #: additional keyword arguments passed to the basin
        self.kwargs = kwargs
        #: Event mapping strategy. If this is "same", it means that the
        #: referring dataset and the basin dataset have identical event
        #: indices. If `mapping` is e.g. `basinmap1` then the mapping of the
        #: indices from the basin to the referring dataset is defined in
        #: `self.basinmap` (copied during initialization of this class from
        #: the array in the key `basinmap1` from the dict-like object
        #: `mapping_referrer`).
        self.mapping = mapping or "same"
        self._basinmap = None  # see `basinmap` property
        # Create a weakref to the original referrer: If it is an instance
        # of RTDCBase, then garbage collection can clean up properly and
        # the basin instance has no reason to exist without the referrer.
        if self.mapping != "same":
            self._basinmap_referrer = weakref.ref(mapping_referrer)
        else:
            self._basinmap_referrer = None
        self._ds = None
        # perform availability check in separate thread
        self._av_check_lock = threading.Lock()
        self._av_check = BasinAvailabilityChecker(self)
        self._av_check.start()

    def __repr__(self):
        options = [
            self.name,
            f"mapped {self.mapping}" if self.mapping != "same" else "",
            f"features {self._features}" if self.features else "full-featured",
            f"location {self.location}",
        ]
        opt_str = ", ".join([o for o in options if o])

        return f"<{self.__class__.__name__} ({opt_str}) at {hex(id(self))}>"

    def _assert_measurement_identifier(self):
        """Make sure the basin matches the measurement identifier

        This method caches its result, i.e. only the first call is slow.
        """
        if not self.verify_basin(run_identifier=True, availability=False):
            raise KeyError(f"Measurement identifier of basin {self.ds} "
                           f"({self.get_measurement_identifier()}) does "
                           f"not match {self.measurement_identifier}!")

    @property
    def basinmap(self):
        """Contains the indexing array in case of a mapped basin"""
        if self._basinmap is None:
            if self.mapping != "same":
                try:
                    basinmap = self._basinmap_referrer()[self.mapping]
                except (KeyError, RecursionError):
                    raise BasinmapFeatureMissingError(
                        f"Could not find the feature '{self.mapping}' in the "
                        f"dataset or any of its basins. This suggests that "
                        f"this feature was never saved anywhere. Please check "
                        f"the input files.")
                #: `basinmap` is an integer array that maps the events from the
                #: basin to the events of the referring dataset.
                self._basinmap = np.array(basinmap,
                                          dtype=np.uint64,
                                          copy=True)
            else:
                self._basinmap = None
        return self._basinmap

    @property
    @abc.abstractmethod
    def basin_format(self):
        """Basin format (:class:`.RTDCBase` subclass), e.g. "hdf5" or "s3"
        """
        # to be implemented in subclasses

    @property
    @abc.abstractmethod
    def basin_type(self):
        """Storage type to use (e.g. "file" or "remote")"""
        # to be implemented in subclasses

    @property
    def ds(self):
        """The :class:`.RTDCBase` instance represented by the basin"""
        if self._ds is None:
            if not self.is_available():
                raise ValueError(f"Basin {self} is not available!")
            self._ds = self.load_dataset(self.location, **self.kwargs)
        return self._ds

    @property
    def features(self):
        """Features made available by the basin

        .. versionchanged: 0.56.0

           Return nested basin features
        """
        if self._features is None:
            if self.is_available():
                # If features are not specified already, either by previous
                # call to this method or during initialization from basin
                # definition, then make the innate and *all* the basin
                # features available.
                self._features = sorted(set(self.ds.features_innate
                                            + self.ds.features_basin))
            else:
                self._features = []
        return self._features

    def as_dict(self):
        """Return basin kwargs for :func:`RTDCWriter.store_basin`

        Note that each subclass of :class:`.RTDCBase` has its own
        implementation of :func:`.RTDCBase.basins_get_dicts` which
        returns a list of basin dictionaries that are used to
        instantiate the basins in :func:`RTDCBase.basins_enable`.
        This method here is only intended for usage with
        :func:`RTDCWriter.store_basin`.
        """
        return {
            "basin_name": self.name,
            "basin_type": self.basin_type,
            "basin_format": self.basin_format,
            "basin_locs": [self.location],
            "basin_descr": self.description,
            "basin_feats": self.features,
            "basin_map": self.basinmap,
        }

    def close(self):
        """Close any open file handles or connections"""
        if self._ds is not None:
            self._ds.close()
        self._av_check.join(0.5)

    def get_feature_data(self, feat):
        """Return an object representing feature data of the basin"""
        self._assert_measurement_identifier()
        return self.ds[feat]

    def get_measurement_identifier(self):
        """Return the identifier of the basin dataset"""
        return self.ds.get_measurement_identifier()

    @abc.abstractmethod
    def is_available(self):
        """Return True if the basin is available"""

    @abc.abstractmethod
    def _load_dataset(self, location, **kwargs):
        """Subclasses should return an instance of :class:`.RTDCBase`"""

    def load_dataset(self, location, **kwargs):
        """Return an instance of :class:`.RTDCBase` for this basin

        If the basin mapping (`self.mapping`) is not the same as
        the referencing dataset
        """
        ds = self._load_dataset(location, **kwargs)
        if self.mapping != "same":
            # apply filter
            ds.filter.manual[:] = False
            ds.filter.manual[self.basinmap] = True
            ds.apply_filter()
            # return hierarchy child
            from .fmt_hierarchy import RTDC_Hierarchy  # avoid circular imports
            ds_bn = RTDC_Hierarchy(ds)
        else:
            ds_bn = ds
        return ds_bn

    def verify_basin(self, availability=True, run_identifier=True):
        if availability:
            check_avail = self.is_available()
        else:
            check_avail = True

        # Only check for run identifier if requested and if the availability
        # check did not fail.
        if run_identifier and check_avail:
            if not self._measurement_identifier_verified:
                if self.measurement_identifier is None:
                    # No measurement identifier was presented by the
                    # referencing dataset. Don't perform any checks.
                    self._measurement_identifier_verified = True
                else:
                    if self.mapping == "same":
                        # When we have identical mapping, then the measurement
                        # identifier has to match exactly.
                        verifier = str.__eq__
                    else:
                        # When we have non-identical mapping (e.g. exported
                        # data), then the measurement identifier has to
                        # partially match.
                        verifier = str.startswith
                    self._measurement_identifier_verified = verifier(
                        self.measurement_identifier,
                        self.get_measurement_identifier()
                    )
            check_rid = self._measurement_identifier_verified
        else:
            check_rid = True

        return check_rid and check_avail


def basin_priority_sorted_key(bdict: Dict):
    """Yield a sorting value for a given basin that can be used with `sorted`

    Basins are normally stored in random order in a dataset. This method
    brings them into correct order, prioritizing:

    - type: "file" over "remote"
    - format: "HTTP" over "S3" over "dcor"
    - mapping: "same" over anything else
    """
    srt_type = {
        "file": "a",
        "remote": "b",
    }.get(bdict.get("type"), "z")

    srt_format = {
        "hdf5": "a",
        "http": "b",
        "s3": "c",
        "dcor": "d",
    }.get(bdict.get("format"), "z")

    mapping = bdict.get("mapping", "same")  # old dicts don't have "mapping"
    srt_map = "a" if mapping == "same" else mapping

    return srt_type + srt_format + srt_map


def get_basin_classes():
    bc = {}
    for b_cls in Basin.__subclasses__():
        if hasattr(b_cls, "basin_format"):
            bc[b_cls.basin_format] = b_cls
    return bc



class H5Tables:
    def __init__(self, h5):
        self.h5file = h5
        self._cache_keys = None

    def __getitem__(self, key):
        if key in self.keys():
            tab = self.h5file["tables"][key]
        else:
            raise KeyError(f"Table '{key}' not found or empty "
                           f"in {self.h5file.file.filename}!")
        return tab

    def __iter__(self):
        # dict-like behavior
        for key in self.keys():
            yield key

    def __len__(self):
        return len(self.keys())

    def keys(self):
        if self._cache_keys is None:
            names = []
            if "tables" in self.h5file:
                for key in self.h5file["tables"]:
                    if self.h5file["tables"][key].size:
                        names.append(key)
            self._cache_keys = names
        return self._cache_keys



import dclab

from helper_methods import example_data_dict


def test_tsv_export(tmp_path):
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=222, keys=keys)
    ds = dclab.new_dataset(ddict)

    f1 = tmp_path / "test.tsv"
    f2 = tmp_path / "test_unicode.tsv"

    ds.export.tsv(f1, keys, override=True)
    ds.export.tsv(f2, [u"area_um", u"deform", u"time",
                       u"frame", u"fl3_width"], override=True)

    with f1.open("r", encoding="utf-8") as fd:
        a1 = fd.read()

    with f2.open("r", encoding="utf-8") as fd:
        a2 = fd.read()

    assert a1 == a2
    assert len(a1) != 0


def test_tsv_override(tmp_path):
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=212, keys=keys)
    ds = dclab.new_dataset(ddict)

    f1 = tmp_path / "test.tsv"
    ds.export.tsv(f1, keys, override=True)
    try:
        ds.export.tsv(f1.with_name(f1.stem), keys, override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .tsv and not override!")


def test_tsv_not_filtered(tmp_path):
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds = dclab.new_dataset(ddict)

    f1 = tmp_path / "test.tsv"
    ds.export.tsv(f1, keys, filtered=False)



"""RT-DC hdf5 format"""
from __future__ import annotations

import pathlib

from .. import feat_basin

from .base import RTDC_HDF5


class HDF5Basin(feat_basin.Basin):
    basin_format = "hdf5"
    basin_type = "file"

    def _load_dataset(self, location, **kwargs):
        return RTDC_HDF5(location, **kwargs)

    def is_available(self):
        with self._av_check_lock:
            avail = False
            try:
                avail = pathlib.Path(self.location).exists()
            except OSError:
                pass
        return avail



"""Tools for linking HDF5 datasets across files"""
from __future__ import annotations

import io
import pathlib
from typing import BinaryIO, Literal

import h5py


class ExternalDataForbiddenError(BaseException):
    """Raised when a dataset contains external data

    External data are a security risk, because they could be
    used to access data that are not supposed to be accessed.
    This is especially critical when the data are accessed within
    a web server process (e.g. in DCOR).
    """
    pass


def assert_no_external(h5):
    """Raise ExternalDataForbiddenError if `h5` refers to external data"""
    has_ext, path_ext = check_external(h5)
    if has_ext:
        raise ExternalDataForbiddenError(
            f"Dataset {h5.file.filename} contains external data, but these "
            f"are not permitted for security reasons ({path_ext})!")


def check_external(h5):
    """Check recursively, whether an h5py object contains external data

    External data includes binary data in external files, virtual
    datasets, and external links.

    Returns a tuple of either

    - `(True, path_ext)` if the object contains external data
    - `(False, None)` if this is not the case

    where `path_ext` is the path to the group or dataset in `h5`.

    .. versionadded:: 0.51.0

    """
    for key in h5:
        obj = h5[key]
        if (obj.file != h5.file  # not in same file
                or (isinstance(obj, h5py.Dataset)
                    and (obj.is_virtual  # virtual dataset
                         or obj.external))):  # external dataset
            # These are external data
            return True, f"{h5.name}/{key}".replace("//", "/")
        elif isinstance(obj, h5py.Group):
            # Perform recursive check for external data
            has_ext, path_ext = check_external(obj)
            if has_ext:
                return True, path_ext
    else:
        return False, None


def combine_h5files(
        paths: list,
        external: Literal["follow", "raise"] = "follow"
        ) -> BinaryIO:
    """Create an in-memory file that combines multiple .rtdc files

    The .rtdc files must have the same number of events. The in-memory
    file is populated with the "events" data from `paths` according to
    the order that `paths` are given in. Metadata, including logs, basins,
    and tables are only taken from the first path.

    .. versionadded:: 0.51.0

    Parameters
    ----------
    paths: list of str or pathlib.Path
        Paths of the input .rtdc files. The first input file is always
        used as a source for the metadata. The other files only complement
        the features.
    external: str
        Defines how external (links, binary, virtual) data in `paths`
        should be handled. The default is to "follow" external datasets or
        links to external data. In a zero-trust context, you can set this
        to "raise" which will cause an :class:`.ExternalDataForbiddenError`
        exception when external data are encountered.

    Returns
    -------
    fd: BinaryIO
        seekable, file-like object representing an HDF5 file opened in
        binary mode; This can be passed to `:class:h5py.File`
    """
    fd = io.BytesIO()
    with h5py.File(fd, "w", libver="latest") as hv:
        for ii, pp in enumerate(paths):
            pp = pathlib.Path(pp).resolve()
            with h5py.File(pp, libver="latest") as h5:
                if external == "raise":
                    # Check for external data
                    assert_no_external(h5)
                if ii == 0:
                    # Only write attributes once.
                    # Interestingly, writing the attributes takes
                    # the most time. Maybe there is some shortcut
                    # that can be taken (since e.g. we know we don't have to
                    # check for existing attributes).
                    # https://github.com/h5py/h5py/blob/master/
                    # h5py/_hl/attrs.py
                    hv.attrs.update(h5.attrs)
                    # Also, write basins/logs/tables/... (anything that is
                    # not events) only once.
                    for group in h5:
                        if group != "events":
                            hv[group] = h5py.ExternalLink(str(pp), group)
                # Append features
                hve = hv.require_group("events")
                for feat in h5["events"]:
                    if feat not in hve:
                        hve[feat] = h5py.ExternalLink(str(pp),
                                                      f"/events/{feat}")
    return fd
